{-
LINGUAGEM DE EXPRESSES 2

MDULO LEXPARSER:
  Definio do Analisador Lxico e da EBNF que produzir a Parse Tree

Equipe:
  Andr Wilson Brotto Furtado
  Gustavo Danzi de Andrade
  Monique Louise Barros Monteiro
-}

module FinalCodeGeneration where

-- TODO: opo do compilador: EXE/DLL
-- TODO: evitar declarao de duas variveis com o mesmo nome em um mesmo let

import InterCodeMonad
import InterCodeMonad
import Semantic

genFinalCode :: String -> InterCodeState -> Type -> InterTerminal -> String
genFinalCode filename ic progType resultLocal =
   (codeHeader filename ic (typeToCode progType)) ++
   localVariables ++
   (foldr (++) [] finalCodeList) ++
   (codeEnd resultLocal)
   where localVariables = initLocalVariables (dataTypes ic)
         finalCodeList = map interElementToCode (codeList ic)

codeHeader :: String -> InterCodeState -> String -> String
codeHeader filename ic progType =
   "// LE2 Compiler - .NET Edition\n\n" ++
   ".assembly " ++ filename ++ " {}\n" ++
   ".namespace LE2 {\n" ++
   "   .class public auto ansi beforefieldinit MainClass extends [mscorlib]System.Object {\n" ++
   "      .method public hidebysig static " ++ progType ++ " AvaliarPrograma() cil managed {\n"

initLocalVariables :: [Type] -> String
initLocalVariables [] = "\n"
initLocalVariables typeList = "      .locals init ( " ++ (initLocalVariablesList typeList 0) ++ " )\n"

initLocalVariablesList :: [Type] -> Int -> String
initLocalVariablesList [] _ = ""
initLocalVariablesList (t:[]) n = initLocalVariable t n
initLocalVariablesList (t:ts) n = (initLocalVariable t n) ++ ", " ++ (initLocalVariablesList ts (n + 1))

initLocalVariable :: Type -> Int -> String
initLocalVariable t n = "[" ++ (show n) ++ "] " ++ (typeToCode t) ++ " rhcil" ++ (show n)

codeEnd :: InterTerminal -> String
codeEnd (TermTemp n) =
          "      ldloc " ++ (show n) ++ "\n" ++
          "      ret\n" ++
          "      }\n" ++
          "   }\n" ++
          "}\n"

terminalToCode :: InterTerminal -> String
terminalToCode (TermInt i) = show i
terminalToCode (TermBool True) = "1"
terminalToCode (TermBool False) = "0"
terminalToCode (TermTemp n) = "rhcil" ++ (show n)

operationToCode :: InterOp -> String
operationToCode (OpMais) = "add"
operationToCode (OpMenos) = "sub"
operationToCode (OpEqual) = "ceq"
operationToCode (OpAnd) = "and"
operationToCode (OpOr) = "or"
operationToCode (OpNot) = "ldc.i4.0" ++ "\n\t" ++ "ceq"
operationToCode (OpConcat) = "call string [mscorlib]System.String::Concat(string,string)"
operationToCode (OpLength) = "callvirt instance int32 [mscorlib]System.String::get_Length()"
operationToCode _ = error "operation not supported \n"

typeToCode :: Type -> String
typeToCode TypeBool = "bool"
typeToCode TypeString = "string"
typeToCode TypeInt = "int32"

--------------------------------------------------------------------------------------
--               TEMPLATES PARA O CODIGO FINAL                     --
--------------------------------------------------------------------------------------

interElementToCode :: InterElement -> String
interElementToCode (AtribBin (TermTemp n) op terminal1 terminal2) =
                   intoCode [
                            loadTerminal terminal1,
                            loadTerminal terminal2,
                            operationToCode op,
                            "stloc " ++ (show n)
                   ]

interElementToCode (AtribUni (TermTemp n) op terminal) =
                   intoCode [
                            loadTerminal terminal,
                            operationToCode op,
                            "stloc " ++ (show n)
                   ]

interElementToCode (Copia (TermTemp n) terminal) =
                  intoCode [
                            loadTerminal terminal,
                            "stloc " ++ (show n)
                   ]

--------------------------------------------------------------------------------------
--               FUNES AUXILIARES                                 --
--------------------------------------------------------------------------------------

loadTerminal :: InterTerminal -> String
loadTerminal (TermInt n) = "ldc.i4." ++ (show n)
loadTerminal (TermBool True) = "ldc.i4.1"
loadTerminal (TermBool False) = "ldc.i4.0"
loadTerminal (TermString str) = "ldstr \"" ++ str ++ "\""
loadTerminal (TermTemp n) = "ldloc " ++ (show n)

------------------------------------------------
--      FUNES DE FORMATAO      --
------------------------------------------------
intoCode :: [String] -> String
intoCode code = foldr (++) [] $ map format code

format :: String -> String
format codeLn = "      " ++ codeLn ++ "\n"


