{-
LINGUAGEM FUNCIONAL 2

MDULO FINALCODEGENERATION:
  Mdulo responsvel pela gerao do cdigo final (MSIL) a partir do cdigo intermedirio

Equipe:
  Andr Wilson Brotto Furtado
  Gustavo Danzi de Andrade
  Monique Louise de Barros Monteiro
-}

module FinalCodeGeneration where

-- TODO: opo do compilador: EXE/DLL
-- TODO: evitar declarao de duas variveis com o mesmo nome em um mesmo let

import InterCodeMonad
import Util

genFinalCode :: String -> InterCodeState -> String
genFinalCode filename ic  =
   (codeHeader filename) ++
   genDelegates (delegates ic) ++
   (foldl (++) [] (map genFunctionsFinalCode (functions ic))) ++
   codeEnd
   where

genFunctionsFinalCode :: FunctionInfo -> String
genFunctionsFinalCode fInfo =
   "\n      .method " ++ visibility ++ " hidebysig static " ++ fType ++ " " ++ fName ++ "(" ++ parameters ++ ")" ++ " cil managed {\n" ++
   localVariables ++
   functionCode ++
   "      ret\n" ++
   "      }\n"
   where fType = show $ returnType fInfo
         fName = name fInfo
         localVariables = initLocalVariables (dataTypes fInfo)
         functionCode = genFunctionCode fInfo (codeList fInfo)
         parameters = parametersToCode (params fInfo)
         visibility | (visible fInfo) = "public"
                    | otherwise = "private"

parametersToCode :: [(String,Bool)] -> String
parametersToCode [] = []
parametersToCode (p:[]) = (parameterToCode p)
parametersToCode (p:ps) = (parameterToCode p) ++ ", " ++ (parametersToCode ps)

parameterToCode :: (String,Bool) -> String
parameterToCode (p,True) = "class [mscorlib]System.MulticastDelegate " ++ p
parameterToCode (p,False) = "object " ++ p

codeHeader :: String -> String
codeHeader filename =
   "// LF2 Compiler - .NET Edition\n\n" ++
   ".assembly " ++ filename ++ " {}\n" ++
   ".namespace LF2 {\n" ++
   "   .class public auto ansi beforefieldinit MainClass extends [mscorlib]System.Object {\n"


initLocalVariables :: [Type] -> String
initLocalVariables [] = "\n"
initLocalVariables typeList = "      .locals init ( " ++ (initLocalVariablesList typeList 0) ++ " )\n"

initLocalVariablesList :: [Type] -> Int -> String
initLocalVariablesList [] _ = ""
initLocalVariablesList (t:[]) n = initLocalVariable t n
initLocalVariablesList (t:ts) n = (initLocalVariable t n) ++ ", " ++ (initLocalVariablesList ts (n + 1))

initLocalVariable :: Type -> Int -> String
initLocalVariable t n = "[" ++ (show n) ++ "] " ++ (typeToCode t) ++ " rhcil" ++ (show n)

codeEnd :: String
codeEnd =
          "   }\n" ++
          "}\n"

terminalToCode :: InterTerminal -> String
terminalToCode (TermInt i) = show i
terminalToCode (TermBool True) = "1"
terminalToCode (TermBool False) = "0"
terminalToCode (TermTemp n) = "rhcil" ++ (show n)

operationToCode :: InterOp -> String
operationToCode (OpMais) = "add"
operationToCode (OpMenos) = "sub"
operationToCode (OpEqual) = "callvirt instance bool [mscorlib]System.Object::Equals(object)"
operationToCode (OpAnd) = "and"
operationToCode (OpOr) = "or"
operationToCode (OpNot) = "ldc.i4.0" ++ "\n\t" ++ "ceq"
operationToCode (OpConcat) = "call string [mscorlib]System.String::Concat(string,string)"
operationToCode (OpLength) = "callvirt instance int32 [mscorlib]System.String::get_Length()"

genDelegates :: [InterTerminal] -> String
genDelegates [] = []
genDelegates ((TermDelegate name (FunctionInfo _ params visible _ returnType _ _ _ _)):ds) =
        "\n   .class auto ansi sealed nested " ++ visibility ++ " " ++ name ++ "_delegate " ++
        " extends [mscorlib]System.MulticastDelegate{\n" ++
        "      .method " ++ visibility ++ " hidebysig specialname rtspecialname " ++
        "instance void  .ctor(object 'object', native int 'method') runtime managed{ }\n" ++
        "      .method " ++ visibility ++ " hidebysig virtual instance " ++ (typeToCode returnType) ++ " Invoke(" ++ parametersToCode params ++
        ") runtime managed { }\n" ++
        "      .method " ++ visibility ++ " hidebysig newslot virtual instance class [mscorlib]System.IAsyncResult \n" ++
        "       BeginInvoke(" ++ formattedParams ++ "class [mscorlib]System.AsyncCallback callback, object 'object') runtime managed { } \n" ++
        "      .method " ++ visibility ++ " hidebysig newslot virtual " ++
        "instance " ++ (typeToCode returnType) ++ " EndInvoke(class [mscorlib]System.IAsyncResult result) runtime managed { }\n" ++
        "\n   }\n" ++
        (genDelegates ds)
        where unFormattedParams = parametersToCode params
              formattedParams | unFormattedParams  == [] = []
                              | otherwise = unFormattedParams ++ ", "
              visibility | visible = "public"
                         | otherwise = "private"

typeToCode :: Type -> String
typeToCode TypeBool = "bool"
typeToCode TypeString = "string"
typeToCode TypeInt = "int32"
typeToCode TypeObject = "object"
typeToCode TypeObjArray = "object[]"
typeToCode TypeUnknown = "FinalCodeGeneration.typeToCode error: TypeUnknown"

--------------------------------------------------------------------------------------
--               TEMPLATES PARA O CODIGO FINAL                     --
--------------------------------------------------------------------------------------

genFunctionCode :: FunctionInfo -> [InterElement] -> String
genFunctionCode _ [] = []
genFunctionCode fInfo (i:is) = (interElementToCode fInfo i) ++ "\n" ++ (genFunctionCode fInfo is)


interElementToCode :: FunctionInfo -> InterElement -> String
interElementToCode fInfo (AtribBin (TermTemp t) (OpEqual) terminal1 terminal2) =
                   intoCode [
                            loadTerminal (terminal1)
                   ] ++
                   makeBox terminal1 fInfo ++
                   intoCode [
                            loadTerminal (terminal2)
                   ] ++
                   makeBox terminal2 fInfo ++
                   intoCode [
                            operationToCode (OpEqual),
                            "stloc " ++ (show t)
                   ]

interElementToCode fInfo (AtribBin (TermTemp t) op terminal1 terminal2) =
                   intoCode [
                            loadTerminal terminal1,
                            makeUnbox terminal1 (getOpType op) fInfo,
                            loadTerminal terminal2,
                            makeUnbox terminal2 (getOpType op) fInfo,
                            operationToCode op,
                            "stloc " ++ (show t)
                   ]

interElementToCode fInfo (AtribUni (TermTemp n) op terminal) =
                   intoCode [
                            loadTerminal terminal,
                            makeUnbox terminal (getOpType op) fInfo,
                            operationToCode op,
                            "stloc " ++ (show n)
                   ]

interElementToCode fInfo (Copia (TermTemp n) terminal) =
                   let copiaType = (getType (dataTypes fInfo) n) in
                   case copiaType of
                      Nothing -> error $ (show (TermTemp n)) ++ " no teve seu tipo declarado!" 
                      (Just TypeObject) -> intoCode [
                                       makeBox terminal fInfo,
                                       loadTerminal terminal,
                                       "stloc " ++ (show n)
                                       ]
                      (Just t) -> intoCode [
                                       makeUnbox terminal t fInfo,
                                       loadTerminal terminal,
                                       "stloc " ++ (show n)
                                       ]
                      
interElementToCode _ (Label name) =
                   intoCode[
                            name ++ ": "
                   ]

interElementToCode fInfo (CreateArray (TermTemp arrayTerminal) size) =
                   intoCode [
                      "ldc.i4 " ++ (show size),
                      "newarr     [mscorlib]System.Object",
                      "stloc " ++  (show arrayTerminal)
                   ]

interElementToCode fInfo (PushArray (TermTemp arrayTerminal) index elemTerminal) =
                   intoCode [
                     "ldloc " ++ (show arrayTerminal),
                     "ldc.i4 " ++ (show index),
                     loadTerminal elemTerminal,
                     "stelem.ref"
                   ]
                   
interElementToCode fInfo (Call (TermCallDelegate orignalName) ps) =
                   intoCode[
                           "callvirt instance object [mscorlib]System.Delegate::DynamicInvoke(object[])"
                   ]

interElementToCode fInfo (Call (TermFunction orignalName returnType) ps) =
                   intoCode[
                            "call " ++  typeToCode(returnType) ++ " LF2.MainClass::" ++
                            orignalName ++ "(" ++ (parameterTypesToCode ps) ++ ")"
                   ]
interElementToCode _ (Pop (TermTemp n)) =
                   intoCode [
                            "stloc " ++ (show n)
                   ]

interElementToCode fInfo (Box terminal) = makeBox terminal fInfo

interElementToCode _ (Push it) =
                   intoCode [loadTerminal it]

interElementToCode _ (GoFalse label) =
                   intoCode [
                            "brfalse " ++ label
                   ]

interElementToCode _ (GoTrue label) =
                   intoCode [
                            "brtrue " ++ label
                   ]

interElementToCode _ (Goto label) =
                   intoCode [
                            "br " ++ label
                   ]

makeBox :: InterTerminal -> FunctionInfo -> String
makeBox (TermInt n) _ = box (Just TypeInt)
makeBox (TermBool b) _ = box (Just TypeBool)
makeBox (TermString str) _ = box (Just TypeString)
makeBox (TermTemp n) fInfo = box(getType (dataTypes fInfo) n)
makeBox (TermArg n) fInfo = "" --  sempre object
makeBox (TermFunction n t) fInfo = "" --  sempre object
makeBox (TermDelegate n f) fInfo = "" --  sempre object
makeBox (TermCallDelegate n) fInfo = error "FinalCodeGeneration.makeBox::TermDelegate no pode ser boxed.\n"

{--
Analisa o tipo passado com parmetro e decide de realiza boxing ou no, dependendo se o tipo
for primitivo ou string.  Caso o tipo seja primitivo, gera o boxing correspondente.
--}
box :: (Maybe Type) -> String
box Nothing = ""
box (Just TypeInt) = intoCode ["box        [mscorlib]System.Int32"]
box (Just TypeBool) = intoCode ["box        [mscorlib]System.Boolean"]
box (Just TypeString) = "" --no faz o boxing, pois j  um objeto
box (Just TypeObject) = "" --no faz o boxing, pois j  um objeto
box (Just TypeUnknown) = error "FinalCodeGeneration.box error: TypeUnknown"

makeUnbox :: InterTerminal -> Type -> FunctionInfo -> String
makeUnbox (TermArg n) t fInfo = unbox t
makeUnbox (TermTemp n) t fInfo | (getType (dataTypes fInfo) n) == (Just TypeObject) = unbox t
                            | otherwise = ""
makeUnbox _ _ _ = ""

{--
Analisa o tipo passado com parmetro e decide de realiza unboxing ou no, dependendo se o tipo
for primitivo ou string.  Caso o tipo seja primitivo, gera o unboxing correspondente.
A instruo "unbox" coloca um ponteiro na pilha, e a instruo "ldind" coloca o valor da variveil associada
ao ponteiro presente na pilha.
--}
unbox :: Type -> String
unbox TypeInt = "unbox        [mscorlib]System.Int32" -- retorna um ponteiro na pilha
                ++ "\n" ++ format "ldind.i4"                             -- retorna na pilha o valor da variavel associada
                                                                       -- ao ponteiro
unbox TypeBool = "unbox        [mscorlib]System.Boolean"
                 ++ "\n" ++ format "ldind.i1"

unbox _ = "" --no faz o unboxing, pois j  um objeto

parameterTypesToCode :: [InterTerminal] -> String
parameterTypesToCode [] = []
parameterTypesToCode (pt:[]) = parameterTypeToCode pt
parameterTypesToCode (pt:pts) = (parameterTypeToCode pt) ++ ", " ++ (parameterTypesToCode pts)

parameterTypeToCode :: InterTerminal -> String
parameterTypeToCode (TermDelegate _ _) = "class [mscorlib]System.MulticastDelegate"
parameterTypeToCode _ = "object"

getType :: [Type] -> Int -> Maybe Type
getType [] _ = Nothing
getType (t:ts) 0 = Just t
getType (t:ts) n = getType ts (n-1)

--------------------------------------------------------------------------------------
--               FUNES AUXILIARES                                 --
--------------------------------------------------------------------------------------

loadTerminal :: InterTerminal -> String
loadTerminal (TermInt n) = "ldc.i4 " ++ (show n)
loadTerminal (TermBool True) = "ldc.i4.1"
loadTerminal (TermBool False) = "ldc.i4.0"
loadTerminal (TermString str) = "ldstr \"" ++ str ++ "\""
loadTerminal (TermTemp n) = "ldloc " ++ (show n)
loadTerminal (TermArg n) = "ldarg " ++ (show n)
loadTerminal (TermDelegate originalName fi) =
                            "ldnull\n      " ++
                            "ldftn      " ++ (typeToCode (returnType fi)) ++ " LF2.MainClass::" ++ -- empilha o ponteiro para a funcao,
                                        originalName ++ "(" ++ (parametersToCode (params fi)) ++ ")\n      " ++
                            "newobj     instance void LF2.MainClass/" ++ originalName ++ "_delegate::.ctor(object,native int)"

------------------------------------------------
--      FUNES DE FORMATAO      --
------------------------------------------------
intoCode :: [String] -> String
intoCode code = foldr (++) [] $ map format code

format :: String -> String
format codeLn = "      " ++ codeLn ++ "\n"
