{-
LINGUAGEM FUNCIONAL 1

MDULO FINALCODEGENERATION:
  Mdulo responsvel pela gerao do cdigo final (MSIL) a partir do cdigo intermedirio

Equipe:
  Andr Wilson Brotto Furtado
  Gustavo Danzi de Andrade
  Monique Louise de Barros Monteiro
-}

module FinalCodeGeneration where

-- TODO: opo do compilador: EXE/DLL
-- TODO: evitar declarao de duas variveis com o mesmo nome em um mesmo let

import InterCodeMonad
import Util

genFinalCode :: String -> InterCodeState -> String
genFinalCode filename ic  =
   (codeHeader filename) ++
   (foldl (++) [] (map genFunctionsFinalCode (functions ic))) ++
   codeEnd
   where

genFunctionsFinalCode :: FunctionInfo -> String
genFunctionsFinalCode fInfo = 
   "      .method public hidebysig static " ++ fType ++ " " ++ fName ++ "(" ++ parameters ++ ")" ++ " cil managed {\n" ++
   localVariables ++ 
   functionCode ++
   
   "      ret\n" ++
   "      }\n"
   where fType = show $ returnType fInfo
         fName = name fInfo
         localVariables = initLocalVariables (dataTypes fInfo)
         functionCode = genFunctionCode fInfo (codeList fInfo)
         parameters = genFunctionParameters (params fInfo)

genFunctionParameters :: [String] -> String
genFunctionParameters [] = []
genFunctionParameters (p:[]) = "object " ++ p
genFunctionParameters (p:ps) = "object " ++ p ++ ", " ++ (genFunctionParameters ps)

codeHeader :: String -> String
codeHeader filename =
   "// LF1 Compiler - .NET Edition\n\n" ++
   ".assembly " ++ filename ++ " {}\n" ++
   ".namespace LF1 {\n" ++
   "   .class public auto ansi beforefieldinit MainClass extends [mscorlib]System.Object {\n"


initLocalVariables :: [Type] -> String
initLocalVariables [] = "\n"
initLocalVariables typeList = "      .locals init ( " ++ (initLocalVariablesList typeList 0) ++ " )\n"

initLocalVariablesList :: [Type] -> Int -> String
initLocalVariablesList [] _ = ""
initLocalVariablesList (t:[]) n = initLocalVariable t n
initLocalVariablesList (t:ts) n = (initLocalVariable t n) ++ ", " ++ (initLocalVariablesList ts (n + 1))

initLocalVariable :: Type -> Int -> String
initLocalVariable t n = "[" ++ (show n) ++ "] " ++ (typeToCode t) ++ " rhcil" ++ (show n)

codeEnd :: String
codeEnd =
          "   }\n" ++
          "}\n"

terminalToCode :: InterTerminal -> String
terminalToCode (TermInt i) = show i
terminalToCode (TermBool True) = "1"
terminalToCode (TermBool False) = "0"
terminalToCode (TermTemp n) = "rhcil" ++ (show n)

operationToCode :: InterOp -> String
operationToCode (OpMais) = "add"
operationToCode (OpMenos) = "sub"
operationToCode (OpEqual) = "callvirt instance bool [mscorlib]System.Object::Equals(object)"
operationToCode (OpAnd) = "and"
operationToCode (OpOr) = "or"
operationToCode (OpNot) = "ldc.i4.0" ++ "\n\t" ++ "ceq"
operationToCode (OpConcat) = "call string [mscorlib]System.String::Concat(string,string)"
operationToCode (OpLength) = "callvirt instance int32 [mscorlib]System.String::get_Length()"

typeToCode :: Type -> String
typeToCode TypeBool = "bool"
typeToCode TypeString = "string"
typeToCode TypeInt = "int32"
typeToCode TypeObject = "object"
typeToCode TypeUnknown = error "FinalCodeGeneration.typeToCode error: TypeUnknown"

--------------------------------------------------------------------------------------
--               TEMPLATES PARA O CODIGO FINAL                     --
--------------------------------------------------------------------------------------

genFunctionCode :: FunctionInfo -> [InterElement] -> String
genFunctionCode _ [] = []
genFunctionCode fInfo (i:is) = (interElementToCode fInfo i) ++ "\n" ++ (genFunctionCode fInfo is)


interElementToCode :: FunctionInfo -> InterElement -> String
interElementToCode fInfo (AtribBin (TermTemp t) (OpEqual) terminal1 terminal2) =
                   intoCode [
                            loadTerminal (terminal1)
                   ] ++
                   makeBox terminal1 fInfo ++
                   intoCode [
                            loadTerminal (terminal2)
                   ] ++
                   makeBox terminal2 fInfo ++
                   intoCode [
                            operationToCode (OpEqual),
                            "stloc " ++ (show t)
                   ]

interElementToCode fInfo (AtribBin (TermTemp t) op terminal1 terminal2) =
                   intoCode [
                            loadTerminal terminal1,
                            makeUnbox terminal1 (getOpType op) fInfo,
                            loadTerminal terminal2,
                            makeUnbox terminal2 (getOpType op) fInfo,
                            operationToCode op,
                            "stloc " ++ (show t)
                   ]

interElementToCode fInfo (AtribUni (TermTemp n) op terminal) =
                   intoCode [
                            loadTerminal terminal,
                            makeUnbox terminal (getOpType op) fInfo,
                            operationToCode op,
                            "stloc " ++ (show n)
                   ]

interElementToCode fInfo (Copia (TermTemp n) terminal) =
                   let copiaType = (getType (dataTypes fInfo) n) in
                   case copiaType of
				      Nothing -> error $ (show (TermTemp n)) ++ " no teve seu tipo declarado!" 
				      (Just TypeObject) -> intoCode [
				                       makeBox terminal fInfo,
				                       loadTerminal terminal,
                                       "stloc " ++ (show n)
                                       ]
				      (Just t) -> intoCode [
                                       makeUnbox terminal t fInfo,
                                       loadTerminal terminal,
                                       "stloc " ++ (show n)
                                       ]
                      
interElementToCode _ (Label name) =
                   intoCode[
                            name ++ ": "
                   ]

interElementToCode fInfo (Call (TermFunction orignalName fType) ps) =
                   intoCode[
                            "call " ++ (show fType) ++ " LF1.MainClass::" ++
                            orignalName ++ "(" ++ (argumentsToCode ps) ++ ")"
                   ]

interElementToCode _ (Pop (TermTemp n)) =
                   intoCode [
                            "stloc " ++ (show n)
                   ]

interElementToCode fInfo (Box terminal) = makeBox terminal fInfo

interElementToCode _ (Push it) =
                   intoCode [loadTerminal it]
                   
interElementToCode _ (GoFalse label) =
                   intoCode [
                            "brfalse " ++ label
                   ]

interElementToCode _ (GoTrue label) =
                   intoCode [
                            "brtrue " ++ label
                   ]

interElementToCode _ (Goto label) =
                   intoCode [
                            "br " ++ label
                   ]

makeBox :: InterTerminal -> FunctionInfo -> String
makeBox (TermInt n) _ = box (Just TypeInt)
makeBox (TermBool b) _ = box (Just TypeBool)
makeBox (TermString str) _ = box (Just TypeString)
makeBox (TermTemp n) fInfo = box(getType (dataTypes fInfo) n)
makeBox (TermArg n) fInfo = "" --  sempre object

{--
Analisa o tipo passado com parmetro e decide de realiza boxing ou no, dependendo se o tipo
for primitivo ou string.  Caso o tipo seja primitivo, gera o boxing correspondente.
--}
box :: (Maybe Type) -> String
box Nothing = ""
box (Just TypeInt) = intoCode ["box        [mscorlib]System.Int32"]
box (Just TypeBool) = intoCode ["box        [mscorlib]System.Boolean"]
box (Just TypeString) = "" --no faz o boxing, pois j  um objeto
box (Just TypeObject) = "" --no faz o boxing, pois j  um objeto
box (Just TypeUnknown) = error "FinalCodeGeneration.box error: TypeUnknown"

makeUnbox :: InterTerminal -> Type -> FunctionInfo -> String
makeUnbox (TermArg n) t fInfo = unbox t
makeUnbox (TermTemp n) t fInfo | (getType (dataTypes fInfo) n) == (Just TypeObject) = unbox t
                            | otherwise = ""
makeUnbox _ _ _ = ""

{--
Analisa o tipo passado com parmetro e decide de realiza unboxing ou no, dependendo se o tipo
for primitivo ou string.  Caso o tipo seja primitivo, gera o unboxing correspondente.
A instruo "unbox" coloca um ponteiro na pilha, e a instruo "ldind" coloca o valor da variveil associada
ao ponteiro presente na pilha.
--}
unbox :: Type -> String
unbox TypeInt = "unbox        [mscorlib]System.Int32" -- retorna um ponteiro na pilha
                ++ "\n" ++ format "ldind.i4"                             -- retorna na pilha o valor da variavel associada
                                                                       -- ao ponteiro
unbox TypeBool = "unbox        [mscorlib]System.Boolean"
				 ++ "\n" ++ format "ldind.i1"

unbox _ = "" --no faz o unboxing, pois j  um objeto

parametersToCode :: [String] -> String
parametersToCode [] = ""
parametersToCode (p:[]) = "object " ++ p
parametersToCode (p:ps) = "object " ++ p ++ ", " ++ (parametersToCode ps)

argumentsToCode :: [InterTerminal] -> String
argumentsToCode [] = ""
argumentsToCode (p:[]) = "object"
argumentsToCode (p:ps) = "object, " ++ (argumentsToCode ps)

getType :: [Type] -> Int -> Maybe Type
getType [] _ = Nothing
getType (t:ts) 0 = Just t
getType (t:ts) n = getType ts (n-1)

--------------------------------------------------------------------------------------
--               FUNES AUXILIARES                                 --
--------------------------------------------------------------------------------------

loadTerminal :: InterTerminal -> String
loadTerminal (TermInt n) = "ldc.i4 " ++ (show n)
loadTerminal (TermBool True) = "ldc.i4.1"
loadTerminal (TermBool False) = "ldc.i4.0"
loadTerminal (TermString str) = "ldstr \"" ++ str ++ "\""
loadTerminal (TermTemp n) = "ldloc " ++ (show n)
loadTerminal (TermArg n) = "ldarg " ++ (show n)

------------------------------------------------
--      FUNES DE FORMATAO      --
------------------------------------------------
intoCode :: [String] -> String
intoCode code = foldr (++) [] $ map format code

format :: String -> String
format codeLn = "      " ++ codeLn ++ "\n"


