{-
LINGUAGEM FUNCIONAL 1

MDULO INTERCODEGENERATION:
  Gerao de Cdigo Intermedirio a partir da Parse Tree

Equipe:
  Andr Wilson Brotto Furtado
  Gustavo Danzi de Andrade
  Monique Louise de Barros Monteiro
-}

module InterCodeGeneration where

import Lexparser
import Util
import InterCodeMonad
import Monad (when,unless)

genInterCode :: Programa -> InterCode ()
genInterCode (Programa expr) = do
                                let newFInfo = FunctionInfo {
                                      name = "AvaliarPrograma", -- TODO: mudar isso caso exista Main
                                      params = [], -- TODO: mudar isso caso exista Main
                                      codeList = [],
                                      returnType = TypeUnknown, -- isso sera' trocado abaixo
                                      nextTemp = 0,
                                      nextArgs = 0,
                                      dataTypes = [],
                                      renameList = []
                                      }
                                
                                addFunctionContext (newFInfo)
                                terminal <- genInterCodeExpression expr
                                fType <- getTypeExpressao expr
                                changeMainFunctionType fType                                
                                addToInterCode (Push terminal)
                                resolveBox terminal fType
                                removeFunctionContext
                                return ()

resolveBox :: InterTerminal -> Type -> InterCode ()
resolveBox terminal TypeObject = addToInterCode (Box terminal)
resolveBox terminal _ = return ()
                                
--------------------------------------------------------------------------------------
--               GERACAO DE CODIGO INTERMEDIARIO DE EXPRESSIONS                     --
--------------------------------------------------------------------------------------

genInterCodeExpression :: Expressao -> InterCode InterTerminal
genInterCodeExpression (ExpressaoValor val) = genInterCodeValor val
genInterCodeExpression (ExpressaoUnaria exp) = genInterCodeExpUnaria exp
genInterCodeExpression (ExpressaoBinaria exp) = genInterCodeExpBinaria exp
genInterCodeExpression (ExpressaoDeclaracao exp) = genInterCodeExpDeclaracao exp
genInterCodeExpression (ExpressaoId ident) = genInterCodeIdent ident
genInterCodeExpression (ExpressaoAplicacao app) = genInterCodeAplicacao app
genInterCodeExpression (ExpressaoIf exp) = genInterCodeIf exp

-- Valor
genInterCodeValor :: Valor -> InterCode InterTerminal
genInterCodeValor (Valor val) = genInterCodeValorConcreto val

-- Valor Concreto
genInterCodeValorConcreto :: ValorConcreto -> InterCode InterTerminal
genInterCodeValorConcreto (ValorInteiro value) = do
          n <- getNextTemp
          addToDataType (TypeInt)
          addToInterCode (Copia (TermTemp n) (TermInt value))
          return (TermTemp n)
genInterCodeValorConcreto (ValorBooleano value) = do
          n <- getNextTemp
          addToDataType (TypeBool)
          addToInterCode (Copia (TermTemp n) (TermBool value))
          return (TermTemp n)
genInterCodeValorConcreto (ValorString value) = do
          n <- getNextTemp
          addToDataType (TypeString)
          addToInterCode (Copia (TermTemp n) (TermString value))
          return (TermTemp n)

-- ExpUnaria
genInterCodeExpUnaria :: ExpUnaria -> InterCode InterTerminal
genInterCodeExpUnaria (PreMenos e) = do
          terminal <- genInterCodeExpression e
          n <- getNextTemp
          addToDataType (TypeInt)
          addToInterCode (AtribUni (TermTemp n) OpMenos terminal)
          return (TermTemp n)
genInterCodeExpUnaria (Not e) = do
          terminal <- genInterCodeExpression e
          n <- getNextTemp
          addToDataType (TypeBool)
          addToInterCode (AtribUni (TermTemp n) OpNot terminal)
          return (TermTemp n)
genInterCodeExpUnaria (Length e) = do
          terminal <- genInterCodeExpression e
          n <- getNextTemp
          addToDataType (TypeString)
          addToInterCode (AtribUni (TermTemp n) OpLength terminal)
          return (TermTemp n)

-- ExpBinaria
genInterCodeExpBinaria :: ExpBinaria -> InterCode InterTerminal
genInterCodeExpBinaria (Mais e1 e2) = do
          
          terminal1 <- genInterCodeExpression e1
          terminal2 <- genInterCodeExpression e2
          n <- getNextTemp
          addToDataType (TypeInt)
          addToInterCode (AtribBin (TermTemp n) OpMais terminal1 terminal2)
          return (TermTemp n)
genInterCodeExpBinaria (Menos e1 e2) = do
          terminal1 <- genInterCodeExpression e1
          terminal2 <- genInterCodeExpression e2
          n <- getNextTemp
          addToDataType (TypeInt)
          addToInterCode (AtribBin (TermTemp n) OpMenos terminal1 terminal2)
          return (TermTemp n)
genInterCodeExpBinaria (And e1 e2) = do
          terminal1 <- genInterCodeExpression e1
          terminal2 <- genInterCodeExpression e2
          n <- getNextTemp
          addToDataType (TypeBool)
          addToInterCode (AtribBin (TermTemp n) OpAnd terminal1 terminal2)
          return (TermTemp n)
genInterCodeExpBinaria (Or e1 e2) = do
          terminal1 <- genInterCodeExpression e1
          terminal2 <- genInterCodeExpression e2
          n <- getNextTemp
          addToDataType (TypeBool)
          addToInterCode (AtribBin (TermTemp n) OpOr terminal1 terminal2)
          return (TermTemp n)
genInterCodeExpBinaria (Igual e1 e2) = do
          terminal1 <- genInterCodeExpression e1
          terminal2 <- genInterCodeExpression e2
          n <- getNextTemp
          addToDataType (TypeBool)
          addToInterCode (AtribBin (TermTemp n) OpEqual terminal1 terminal2)
          return (TermTemp n)
genInterCodeExpBinaria (Concatenacao e1 e2) = do
          terminal1 <- genInterCodeExpression e1
          terminal2 <- genInterCodeExpression e2
          n <- getNextTemp
          addToDataType (TypeString)
          addToInterCode (AtribBin (TermTemp n) OpConcat terminal1 terminal2)
          return (TermTemp n)

-- ExpDeclaracao
genInterCodeExpDeclaracao :: ExpDeclaracao -> InterCode InterTerminal
genInterCodeExpDeclaracao (LetIn decFuncional exp) = do
          genInterCodeDecFuncional decFuncional
          genInterCodeExpression exp

genInterCodeDecFuncional :: DeclaracaoFuncional -> InterCode ()
genInterCodeDecFuncional (Var ident exp) = do
          terminal <- genInterCodeExpression exp
          n <- getNextTemp
          expType <- getTypeExpressao exp
          addToDataType expType
          addToRenameList (ident, (TermTemp n))
          addToInterCode (Copia (TermTemp n) (terminal))
genInterCodeDecFuncional (Fun fName fParams fBody) = do
                                        let newFInfo = FunctionInfo {
                                              name = fName,
                                              params = fParams,
                                              codeList = [],
                                              returnType = TypeUnknown,
                                              nextTemp = 0,
                                              nextArgs = 0,
                                              dataTypes = [],
                                              renameList = []
                                            }
                                        addFunctionContext(newFInfo)
                                        fType <- getTypeExpressao fBody
                                        changeFunctionType(fType)
                                        addToRenameFunctionsList (fName, (TermFunction fName fType))
                                        addParametersToRenameList fParams
                                        terminal <- genInterCodeExpression fBody
                                        addToInterCode (Push terminal)
                                        removeFunctionContext
                                        return ()

genInterCodeDecFuncional (FuncList dec1 dec2) = do
                                        genInterCodeDecFuncional dec1
                                        genInterCodeDecFuncional dec2


-- Id
genInterCodeIdent :: String -> InterCode InterTerminal
genInterCodeIdent ident = getFromRenameList ident

-- Aplicao de Funo
genInterCodeAplicacao :: ExpAplicacao -> InterCode InterTerminal
genInterCodeAplicacao (Aplicacao fName expList) = do
                                        params <- genInterCodeParams expList
                                        terminal <- getFromRenameFunctionsList fName
                                        addToInterCode (Call terminal params)
                                        newTerminal <- getNextTemp
                                        functionInfo <- getFunctionInfo fName
                                        addToDataType (returnType functionInfo)
                                        addToInterCode (Pop (TermTemp newTerminal))
                                        return (TermTemp newTerminal)


genInterCodeParams :: [Expressao] -> InterCode [InterTerminal]
genInterCodeParams expList = genInterCodeParamsAux [] expList

genInterCodeParamsAux :: [InterTerminal] -> [Expressao] -> InterCode [InterTerminal]
genInterCodeParamsAux accum [] = return $ reverse accum
genInterCodeParamsAux accum (e:es) = do
                                        terminal <- genInterCodeExpression e
                                        addToInterCode (Push terminal)
                                        addToInterCode (Box terminal)
                                        genInterCodeParamsAux (terminal:accum) es


-- IfThenElse
genInterCodeIf :: ExpIf -> InterCode InterTerminal
genInterCodeIf (IfThenElse guarda expThen expElse) = do
  nextTemp <- getNextTemp
  let terminal = TermTemp nextTemp
  expType <- (getTypeExpressao expThen)
  addToDataType expType
  labelLeaveIf <- getNextLabel
  labelElse <- getNextLabel
  guardaTerminal <- genInterCodeExpression guarda
  addToInterCode(Push guardaTerminal)
  addToInterCode (GoFalse labelElse)
  expThenTerminal <- genInterCodeExpression expThen
  addToInterCode (Copia terminal expThenTerminal)
  addToInterCode (Goto labelLeaveIf)
  addToInterCode(Label labelElse)
  expElseTerminal <- genInterCodeExpression expElse
  addToInterCode (Copia terminal expElseTerminal)
  addToInterCode(Label labelLeaveIf)
  return terminal

{-


-- IfThenElse: VERSO COM OTIMIZAES (mlbm)
genInterCodeIf :: ExpIf -> InterCode InterTerminal
genInterCodeIf (IfThenElse guarda expThen expElse) = do
  nextTemp <- getNextTemp
  let terminal = TermTemp nextTemp
  expType <- (getTypeExpressao expThen)
  addToDataType expType
  labelIf <- getNextLabel
  labelLeaveIf <- getNextLabel
  labelElse <- getNextLabel
  genConditionalBranch guarda True labelIf labelElse
  addToInterCode (Label labelIf)
  expThenTerminal <- genInterCodeExpression expThen
  addToInterCode (Copia terminal expThenTerminal)
  addToInterCode (Goto labelLeaveIf)
  addToInterCode(Label labelElse)
  expElseTerminal <- genInterCodeExpression expElse
  addToInterCode (Copia terminal expElseTerminal)
  addToInterCode(Label labelLeaveIf)
  return terminal


------------------------ FUNES ESPECIAIS PARA GERAO DE COMANDOS DE DESVIO ---------------

genConditionalBranch :: Expressao -> Bool -> String -> String -> InterCode ()
genConditionalBranch (ExpressaoBinaria exp) _ labelTrue labelFalse = genConditionalBranchExpBinaria exp labelTrue labelFalse
genConditionalBranch (ExpressaoId ident) isTopLevel labelTrue labelFalse = genConditionalBranchExpIdent ident isTopLevel labelTrue labelFalse
genConditionalBranch (ExpressaoValor val) isTopLevel labelTrue labelFalse = genConditionalBranchExpValor val isTopLevel labelTrue labelFalse
genConditionalBranch (ExpressaoUnaria exp) isTopLevel labelTrue labelFalse = genConditionalBranchExpUnaria exp isTopLevel labelTrue labelFalse

genConditionalBranchExpUnaria :: ExpUnaria -> Bool -> String -> String -> InterCode ()
genConditionalBranchExpUnaria (Not e) isTopLevel labelTrue labelFalse = do
        terminal <- genConditionalBranch e isTopLevel labelFalse labelTrue
        return ()

genConditionalBranchExpBinaria :: ExpBinaria -> String -> String -> InterCode ()
genConditionalBranchExpBinaria (Igual e1 e2) labelTrue labelFalse = do
                     terminal <- genInterCodeExpBinaria (Igual e1 e2)                     
                     addToInterCode (GoFalse labelFalse)
genConditionalBranchExpBinaria (And e1 e2) labelTrue labelFalse = do
                     tempLabel <- getNextLabel
                     genConditionalBranch e1 False tempLabel labelFalse
                     addToInterCode (GoFalse labelFalse)
                     addToInterCode (Label tempLabel)
                     genConditionalBranch e2 False labelTrue labelFalse
                     addToInterCode (GoFalse labelFalse)
genConditionalBranchExpBinaria (Or e1 e2) labelTrue labelFalse = do
                     tempLabel <- getNextLabel
                     genConditionalBranch e1 False labelTrue tempLabel
                     addToInterCode (Label tempLabel)
                     addToInterCode (GoTrue labelTrue)
                     genConditionalBranch e2 False labelTrue labelFalse
                     addToInterCode (GoFalse labelFalse)

genConditionalBranchExpValor :: Valor -> Bool -> String -> String -> InterCode ()
genConditionalBranchExpValor v isTopLevel labelTrue labelFalse =
        case isTopLevel of
             True -> do
                       terminal <- genInterCodeExpression (ExpressaoValor v)
                       addToInterCode (GoFalse labelFalse)
             False -> do
                       terminal <- genInterCodeExpression (ExpressaoValor v)
                       return ()


genConditionalBranchExpIdent :: String -> Bool -> String -> String -> InterCode ()
genConditionalBranchExpIdent ident isTopLevel labelTrue labelFalse =
        case isTopLevel of
            True -> do
                       terminal <- genInterCodeExpression (ExpressaoId ident)
                       addToInterCode (GoFalse labelFalse)
            False -> do
                       terminal <- genInterCodeExpression (ExpressaoId ident)
                       return ()
-}