{-

MDULO INTERCODEGENERATION:
	Gerao de Cdigo Intermedirio a partir da Parse Tree

Equipe:
	Andr Wilson Brotto Furtado
	Gustavo Danzi de Andrade
	Monique Luise Barros Monteiro
-}

module InterCodeGeneration where

import Lexparser
import Semantic
import InterCodeMonad
import Monad (when,unless)

genInterCode :: Programa -> InterCode InterTerminal
genInterCode (Programa expr) = genInterCodeExpressao expr

--------------------------------------------------------------------------------------
--               GERACAO DE CODIGO INTERMEDIARIO DE EXPRESSIONS                     --
--------------------------------------------------------------------------------------

genInterCodeExpressao :: Expressao -> InterCode InterTerminal
genInterCodeExpressao (ExpressaoValor val) = genInterCodeValor val
genInterCodeExpressao (ExpressaoUnaria exp) = genInterCodeExpUnaria exp
genInterCodeExpressao (ExpressaoBinaria exp) = genInterCodeExpBinaria exp
genInterCodeExpressao (ExpressaoDeclaracao exp) = genInterCodeExpDeclaracao exp
genInterCodeExpressao (ExpressaoId ident) = genInterCodeIdent ident

-- Valor
genInterCodeValor :: Valor -> InterCode InterTerminal
genInterCodeValor (Valor val) = genInterCodeValorConcreto val

-- Valor Concreto
genInterCodeValorConcreto :: ValorConcreto -> InterCode InterTerminal
genInterCodeValorConcreto (ValorInteiro value) = do
					n <- getNextTemp
					addToDataType (TypeInt)
					addToInterCode (Copia (TermTemp n) (TermInt value))
					return (TermTemp n)
genInterCodeValorConcreto (ValorBooleano value) = do
					n <- getNextTemp
					addToDataType (TypeBool)
					addToInterCode (Copia (TermTemp n) (TermBool value))
					return (TermTemp n)
genInterCodeValorConcreto (ValorString value) = do
					n <- getNextTemp
					addToDataType (TypeString)
					addToInterCode (Copia (TermTemp n) (TermString value))
					return (TermTemp n)

-- ExpUnaria
genInterCodeExpUnaria :: ExpUnaria -> InterCode InterTerminal
genInterCodeExpUnaria (PreMenos e) = do
					terminal <- genInterCodeExpressao e
					n <- getNextTemp
					addToDataType (TypeInt)
					addToInterCode (AtribUni (TermTemp n) OpMenos terminal)
					return (TermTemp n)
genInterCodeExpUnaria (Not e) = do
					terminal <- genInterCodeExpressao e
					n <- getNextTemp
					addToDataType (TypeBool)
					addToInterCode (AtribUni (TermTemp n) OpNot terminal)
					return (TermTemp n)
genInterCodeExpUnaria (Length e) = do
					terminal <- genInterCodeExpressao e
					n <- getNextTemp
					addToDataType (TypeString)
					addToInterCode (AtribUni (TermTemp n) OpLength terminal)
					return (TermTemp n)

-- ExpBinaria
genInterCodeExpBinaria :: ExpBinaria -> InterCode InterTerminal
genInterCodeExpBinaria (Mais e1 e2) = do
					terminal1 <- genInterCodeExpressao e1
					terminal2 <- genInterCodeExpressao e2
					n <- getNextTemp
					addToDataType (TypeInt)
					addToInterCode (AtribBin (TermTemp n) OpMais terminal1 terminal2)
					return (TermTemp n)
genInterCodeExpBinaria (Menos e1 e2) = do
					terminal1 <- genInterCodeExpressao e1
					terminal2 <- genInterCodeExpressao e2
					n <- getNextTemp
					addToDataType (TypeInt)
					addToInterCode (AtribBin (TermTemp n) OpMenos terminal1 terminal2)
					return (TermTemp n)
genInterCodeExpBinaria (And e1 e2) = do
					terminal1 <- genInterCodeExpressao e1
					terminal2 <- genInterCodeExpressao e2
					n <- getNextTemp
					addToDataType (TypeBool)
					addToInterCode (AtribBin (TermTemp n) OpAnd terminal1 terminal2)
					return (TermTemp n)
genInterCodeExpBinaria (Or e1 e2) = do
					terminal1 <- genInterCodeExpressao e1
					terminal2 <- genInterCodeExpressao e2
					n <- getNextTemp
					addToDataType (TypeBool)
					addToInterCode (AtribBin (TermTemp n) OpOr terminal1 terminal2)
					return (TermTemp n)
genInterCodeExpBinaria (Igual e1 e2) = do
					terminal1 <- genInterCodeExpressao e1
					terminal2 <- genInterCodeExpressao e2
					n <- getNextTemp
					addToDataType (TypeBool)
					addToInterCode (AtribBin (TermTemp n) OpEqual terminal1 terminal2)
					return (TermTemp n)
genInterCodeExpBinaria (Concatenacao e1 e2) = do
					terminal1 <- genInterCodeExpressao e1
					terminal2 <- genInterCodeExpressao e2
					n <- getNextTemp
					addToDataType (TypeString)
					addToInterCode (AtribBin (TermTemp n) OpConcat terminal1 terminal2)
					return (TermTemp n)

-- ExpDeclaracao
genInterCodeExpDeclaracao :: ExpDeclaracao -> InterCode InterTerminal
genInterCodeExpDeclaracao (LetIn decVars exp) = do
					genInterCodeDecVariavel decVars
					genInterCodeExpressao exp

genInterCodeDecVariavel :: DecVariavel -> InterCode ()
genInterCodeDecVariavel (VarList d1 d2) = do
					genInterCodeDecVariavel d1
					genInterCodeDecVariavel d2
genInterCodeDecVariavel (Var ident exp) = do
					terminal <- genInterCodeExpressao exp
					n <- getNextTemp
					addToDataType (getTypeExpressao exp)
					addToRenameList (ident, (TermTemp n))
					addToInterCode (Copia (TermTemp n) (terminal))

-- Id
genInterCodeIdent :: String -> InterCode InterTerminal
genInterCodeIdent ident = getFromRenameList ident


