{-
MDULO INTERCODEMONAD:
	Mdulo com a definio e as operaes sobre o Monad InterCode

Equipe:
	Andr Wilson Brotto Furtado
	Gustavo Danzi de Andrade
	Monique Luise Barros Monteiro
-}

-- definio e funes mondicas para o Monad InterCode
module InterCodeMonad where

import Semantic

-- Os tipos de ns da rvore de cdigo intermedirio
-- elementos nao terminais...
data InterElement
	= AtribBin InterTerminal InterOp InterTerminal InterTerminal
    | AtribUni InterTerminal InterOp InterTerminal
    | Copia InterTerminal InterTerminal

-- elementos terminais...
data InterOp
	     = OpMais
	     | OpMenos
	     | OpEqual
         | OpAnd
 	     | OpOr
	     | OpNot
	     | OpConcat
	     | OpLength

data InterTerminal
	     = TermInt Int
	     | TermBool Bool
	     | TermString String
--	     | TermIdent String
	     | TermTemp Int

-- Implementao do show dos tipos de dados do cdigo intermedirio
instance  Show InterElement where
	show (AtribBin t1 op t2 t3 	) = "    " ++ show(t1) ++ ":=" ++ show(t2) ++ " " ++ show(op) ++ " " ++ show(t3)
	show (AtribUni t1 op t2		) = "    " ++ show(t1) ++ ":=" ++ " " ++ show(op) ++ " " ++ show(t2)
	show (Copia t1 t2		) = "    " ++ show(t1) ++ ":=" ++ show(t2)

instance  Show InterOp where
	show (OpMais		) = "+"
	show (OpMenos		) = "-"
	show (OpEqual		) = "="
	show (OpAnd		    ) = "AND"
	show (OpOr		    ) = "OR"
	show (OpNot		    ) = "NOT"
	show (OpConcat		) = "++"
	show (OpLength	    ) = "LENGTH"

instance  Show InterTerminal where
	show (TermInt	 i) = show(i)
	show (TermBool	 b) = show(b)
	show (TermString s) = s
-- 	show (TermIdent  s) = s
	show (TermTemp   n) = "rhcil" ++ show(n)

data InterCodeState = InterCodeState {
	nextTemp      :: Int,
	dataTypes     :: [Type],
--	nextLabel     :: Int,
	codeList      :: [InterElement],
	renameList    :: [(String,InterTerminal)]     -- associa um ident  um TermTemp
--	identList     :: IdentList,
--	interTypeList :: TypeList

	}

-- Definies padres de Monads
data InterCode a = IC ( InterCodeState -> (InterCodeState, a) )

instance Monad InterCode
  where
    -- (>>=) :: InterCodeState a -> (a -> InterCodeState b) -> InterCodeState b
    (IC p) >>= k  =  IC( \s0 -> let (s1, a) = p s0
                                    (IC q) = k a
                                in q s1 )

    -- return :: a -> InterCodeState a
    return a = IC( \interCode -> (interCode, a) )

---------------------------------------------------------------------------------
--                             FUNES MONDICAS                               --
---------------------------------------------------------------------------------
-- funcao de aplicacao do Monad
applyInterCode :: InterCode InterTerminal -> InterCodeState -> (InterCodeState,InterTerminal)
applyInterCode (IC f) interCodeState = f interCodeState

getNextTemp :: InterCode Int
getNextTemp = IC(\interCode -> (interCode {nextTemp = (nextTemp interCode) + 1}, nextTemp interCode))

addToDataType :: Type -> InterCode ()
addToDataType typeElement = IC(\interCode -> (interCode {dataTypes = (dataTypes interCode) ++ [typeElement]},()))

addToInterCode :: InterElement -> InterCode ()
addToInterCode interElement = IC(\interCode -> (interCode {codeList = (codeList interCode) ++ [interElement]},()))

addToRenameList :: (String, InterTerminal) -> InterCode ()
addToRenameList renamedElement = IC(\interCode -> (interCode {renameList = (renameList interCode) ++ [renamedElement]},()))

getFromRenameList :: String -> InterCode InterTerminal
getFromRenameList ident = IC(\interCode -> (interCode, getFromRenameListAux ident (renameList interCode)))

getFromRenameListAux :: String -> [(String,InterTerminal)] -> InterTerminal
getFromRenameListAux ident [] = error ("getFromRenameListAux.error: Nao foi encontrado nenhum ident de nome " ++ ident)
getFromRenameListAux ident ((identName, identTerm):is) | (ident == identName) = identTerm
			 | otherwise = getFromRenameListAux ident is





