LINGUAGEM FUNCIONAL 1

MDULO LEXPARSER:
  Definio do Analisador Lxico e da EBNF que produzir a Parse Tree

Equipe:
  Andr Wilson Brotto Furtado
  Gustavo Danzi de Andrade
  Monique Louise de Barros Monteiro

> {
> module Lexparser where
> import Char
> }

Nome do parser e declaraao do tipo algebrico dos Tokens.

> %name parserLF1
> %tokentype { Token }

----------------------------------
Declaraao de Tokens para o Parser
----------------------------------

> %token
>       valorInt      {TokenInt $$}
>       valorBool     {TokenBool $$}
>       valorString   {TokenString $$}
>       not           {TokenNot $$}
>       length        {TokenLength $$}
>       '+'           {TokenMais $$}
>       '-'           {TokenMenos $$}
>       and           {TokenAnd $$}
>       or            {TokenOr $$}
>       '=='          {TokenIgual $$}
>       '++'          {TokenConcatenacao $$}
>       'let'         {TokenLet $$}
>       'in'          {TokenIn $$}
>       'var'         {TokenVar $$}
>       ident         {TokenId $$}
>       '='           {TokenAtribuicao $$}
>       ','           {TokenVirgula $$}
>       'if'          {TokenIf $$}
>       'then'        {TokenThen $$}
>       'else'        {TokenElse $$}
>       'fun'         {TokenFun $$}
>       '('           {TokenLeftPar $$}
>       ')'           {TokenRightPar $$}

The left hand side are the names of the terminals or tokens,
and the right hand side is how to pattern match them.

> %nonassoc '=='
> %right not length
> %left '++'
> %left '+' '-' or
> %left and
> %%

---------------------------------
Now we have the production rules.
---------------------------------

> Programa :: { Programa }
> Programa : Expressao                              {Programa $1}

> Expressao :: { Expressao }
> Expressao : Valor                                 {ExpressaoValor $1}
>           | ExpUnaria                             {ExpressaoUnaria $1}
>           | ExpBinaria                            {ExpressaoBinaria $1}
>           | ExpDeclaracao                         {ExpressaoDeclaracao $1}
>           | ident                                 {ExpressaoId (fst $1)}
>           | ExpAplicacao                          {ExpressaoAplicacao $1}
>           | ExpIf                                 {ExpressaoIf $1}

> Valor :: { Valor }
> Valor : ValorConcreto                             {Valor $1}

> ValorConcreto :: { ValorConcreto }
> ValorConcreto : valorInt                          {ValorInteiro (fst $1)}
>               | valorBool                         {ValorBooleano (fst $1)}
>               | valorString                       {ValorString (fst $1)}

> ExpUnaria :: { ExpUnaria }
> ExpUnaria : '-' Expressao                         {PreMenos $2}
>           | not Expressao                         {Not $2}
>           | length Expressao                      {Length $2}

> ExpBinaria :: { ExpBinaria }
> ExpBinaria : Expressao '+' Expressao              {Mais $1 $3}
>            | Expressao and Expressao              {And $1 $3}
>            | Expressao '-' Expressao              {Menos $1 $3}
>            | Expressao or Expressao               {Or $1 $3}
>            | Expressao '==' Expressao             {Igual $1 $3}
>            | Expressao '++' Expressao             {Concatenacao $1 $3}

> ExpDeclaracao :: { ExpDeclaracao }
> ExpDeclaracao : 'let' DeclaracaoFuncional 'in' Expressao  {LetIn $2 $4}

> DeclaracaoFuncional :: {DeclaracaoFuncional}
> DeclaracaoFuncional : DeclaracaoFuncionalTerminal   {$1}
>                     | DeclaracaoFuncional ',' DeclaracaoFuncionalTerminal {FuncList $1 $3}

> DeclaracaoFuncionalTerminal :: {DeclaracaoFuncional}
> DeclaracaoFuncionalTerminal : DecVariavel         {$1}
>                             | DecFuncao           {$1}

> DecVariavel :: { DeclaracaoFuncional }
> DecVariavel : 'var' ident '=' Expressao           {Var (fst $2) $4}

> DecFuncao :: { DeclaracaoFuncional }
> DecFuncao : 'fun' ident IdentList '=' Expressao {Fun (fst $2) $3 $5}

> IdentList :: {[String]}
> IdentList : ident IdentList {(fst $1):$2}
>           | {- vazio -}     {[]}

> ExpAplicacao :: { ExpAplicacao }
> ExpAplicacao : ident '(' ListaArgumentos ')' {Aplicacao (fst $1) $3}

> ListaArgumentos :: {[Expressao]}
> ListaArgumentos : ListaArgumentosNaoVazia            {$1}
>                 | {- vazio -}                        {[]}

> ListaArgumentosNaoVazia :: {[Expressao]}
> ListaArgumentosNaoVazia : Expressao                                {[$1]}
>                         | ListaArgumentosNaoVazia ',' Expressao    {$1 ++ [$3]}

> ExpIf :: { ExpIf }
> ExpIf : 'if' Expressao 'then' Expressao 'else' Expressao {IfThenElse $2 $4 $6}

We are simply returning the parsed data structure !
Now we need some extra code, to support this parser,
and make in complete:

> {

All parsers must declair this function,
which is called when an error is detected.
Note that currently we do no error recovery.

> happyError :: [Token] -> a
> happyError [] = error "Erro sintatico na ultima linha."
> happyError (t:_)= error (getTokenMessage t)

> getTokenMessage :: Token -> String
> getTokenMessage (TokenId (t,n)) = "Erro sintatico (linha " ++ (show n)  ++ "): proximo ao ident " ++ t ++ "."
> getTokenMessage (TokenInt (t,n)) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao int " ++ show(t) ++ "."
> getTokenMessage (TokenBool (t,n)) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao bool " ++ show(t) ++ "."
> getTokenMessage (TokenString (t,n)) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao string " ++ show(t) ++ "."
> getTokenMessage (TokenNot n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao operador logico not."
> getTokenMessage (TokenLength n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao termo length."
> getTokenMessage (TokenMais n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao operador aritmetico +."
> getTokenMessage (TokenMenos n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao operador aritmetico -."
> getTokenMessage (TokenAnd n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao operador logico and."
> getTokenMessage (TokenOr n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao operador logico or ."
> getTokenMessage (TokenIgual n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao operador logico ==."
> getTokenMessage (TokenConcatenacao n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao termo ++."
> getTokenMessage (TokenLet n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao termo let."
> getTokenMessage (TokenIn n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao termo in."
> getTokenMessage (TokenVar n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao termo var."
> getTokenMessage (TokenAtribuicao n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo a operador = ."
> getTokenMessage (TokenVirgula n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo a ',' ."
> getTokenMessage (TokenIf n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao termo if."
> getTokenMessage (TokenThen n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao termo then."
> getTokenMessage (TokenElse n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo ao termo else."
> getTokenMessage (TokenLeftPar n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo a '('."
> getTokenMessage (TokenRightPar n) =  "Erro sintatico (linha " ++ (show n) ++ "): proximo a ')'."

Now we declare the datastructure that we are parsing.

> data Programa = Programa Expressao
>                 deriving (Show)
>
> data Expressao = ExpressaoValor Valor
>                | ExpressaoUnaria ExpUnaria
>                | ExpressaoBinaria ExpBinaria
>                | ExpressaoDeclaracao ExpDeclaracao
>                | ExpressaoId String
>                | ExpressaoAplicacao ExpAplicacao
>                | ExpressaoIf ExpIf
>                  deriving (Show)
>
> data Valor = Valor ValorConcreto
>              deriving (Show)
>
> data ValorConcreto = ValorInteiro Int
>                    | ValorBooleano Bool
>                    | ValorString String
>                      deriving (Show)
>
> data ExpUnaria = PreMenos Expressao
>                | Not Expressao
>                | Length Expressao
>                  deriving (Show)
>
> data ExpBinaria = Mais Expressao Expressao
>                 | Menos Expressao Expressao
>                 | And Expressao Expressao
>                 | Or Expressao Expressao
>                 | Igual Expressao Expressao
>                 | Concatenacao Expressao Expressao
>                   deriving (Show)
>
> data ExpDeclaracao = LetIn DeclaracaoFuncional Expressao
>                      deriving (Show)
>
> data DeclaracaoFuncional = Var String Expressao
>                          | Fun String [String] Expressao
>                          | FuncList DeclaracaoFuncional DeclaracaoFuncional
>                          deriving (Show)
>
> data ExpAplicacao = Aplicacao String [Expressao]
>                     deriving (Show)
>
> data ExpIf = IfThenElse Expressao Expressao Expressao
>              deriving (Show)

-------------------------------------------------
O tipo algebrico definindo os tokens do programa
-------------------------------------------------

> data Token
>   = TokenInt (Int,Int) -- o segundo elemento e' a linha
>   | TokenBool (Bool,Int)
>   | TokenString (String,Int)
>   | TokenNot Int
>   | TokenLength Int
>   | TokenMais Int
>   | TokenMenos Int
>   | TokenAnd Int
>   | TokenOr Int
>   | TokenIgual Int
>   | TokenConcatenacao Int
>   | TokenLet Int
>   | TokenIn Int
>   | TokenVar Int
>   | TokenId (String,Int)
>   | TokenAtribuicao Int
>   | TokenVirgula Int
>   | TokenIf Int
>   | TokenThen Int
>   | TokenElse Int
>   | TokenFun Int
>   | TokenLeftPar Int
>   | TokenRightPar Int
>   deriving Show

----------------------------------------------
A funao lexica que cuida desse tipo algebrico
----------------------------------------------

> lexInit :: String -> [Token]
> lexInit file = lexer file 1

> lexer :: String -> Int -> [Token]
> lexer [] _ = []
> lexer ('\n':cs) ct = lexer cs (ct + 1)
> lexer (c:cs) ct
>   | isSpace c = lexer cs ct
>   | isAlpha c = lexAlpha (c:cs) ct
>   | isDigit c = lexNum (c:cs) ct
> lexer ('/':cs) ct = lexBarras cs ct
> lexer ('{':cs) ct = lexComentario cs ct
> lexer ('=':cs) ct = lexIgual cs ct
> lexer (',':cs) ct = TokenVirgula ct : lexer cs ct
> lexer ('+':cs) ct = lexMais cs ct
> lexer ('-':cs) ct = TokenMenos ct : lexer cs ct
> lexer ('(':cs) ct = TokenLeftPar ct : lexer cs ct
> lexer (')':cs) ct = TokenRightPar ct : lexer cs ct
> lexer ('\"':cs) ct = lexString cs ct
> lexer (e  :cs) ct = error (" Erro lexico (linha " ++ show(ct) ++ "): caractere invalido: " ++ [e])

> ignoreLine ('\n':cs) ct = lexer cs (ct+1)
> ignoreLine (_:cs) ct = ignoreLine cs ct
>
> lexBarras ('/':cs) ct = ignoreLine cs ct
> lexBarras _ ct = error (" Erro lexico (linha " ++ show(ct) ++ "): \'/\' inesperado!")
>
> lexComentario ('}':cs) ct = lexer cs ct
> lexComentario ('\n':cs) ct = lexComentario cs (ct+1)
> lexComentario (_:cs) ct = lexComentario cs ct

> lexIgual ('=':cs) ct = TokenIgual ct      : lexer cs ct
> lexIgual (_:cs)   ct = TokenAtribuicao ct : lexer cs ct

> lexMais  ('+':cs) ct = TokenConcatenacao ct : lexer cs ct
> lexMais (_:cs)   ct = TokenMais ct         : lexer cs ct

> lexNum cs ct = (TokenInt (read num,ct)) : lexer rest ct
>    where (num,rest) = span isDigit cs
>
> lexString cs ct = (TokenString (str, ct))   : lexer rest ct
>           where (str, r:rest) = span isAspas cs
>
> lexAlpha cs ct =
>    let (token,rest) = span isAlphaOrDigit cs
>    in case (map toUpper token) of
>           "NOT"       -> TokenNot ct          : lexer rest ct
>           "LENGTH"    -> TokenLength ct       : lexer rest ct
>           "AND"       -> TokenAnd ct          : lexer rest ct
>           "OR"        -> TokenOr ct           : lexer rest ct
>           "LET"       -> TokenLet ct          : lexer rest ct
>           "IN"        -> TokenIn ct           : lexer rest ct
>           "VAR"       -> TokenVar ct          : lexer rest ct
>           "TRUE"      -> TokenBool (True,ct)  : lexer rest ct
>           "FALSE"     -> TokenBool (False,ct) : lexer rest ct
>           "IF"        -> TokenIf ct           : lexer rest ct
>           "THEN"      -> TokenThen ct         : lexer rest ct
>           "ELSE"      -> TokenElse ct         : lexer rest ct
>           "FUN"       -> TokenFun ct          : lexer rest ct
>           "LEFT_PAR"  -> TokenLeftPar ct      : lexer rest ct
>           "RIGHT_PAR" -> TokenRightPar ct     : lexer rest ct
>           _           -> TokenId (token,ct)   : lexer rest ct

> isAspas :: Char -> Bool
> isAspas = \x -> x /= '\"'

> isAlphaOrDigit :: Char -> Bool
> isAlphaOrDigit c = (isAlpha c) || (isDigit c)

------------------------------------------------------
               Funes de debugagem
------------------------------------------------------

> -- usada apenas para debugar a parte lexica
> -- faz a analise lexica de um arquivo, dado o nome do mesmo
> runLexer :: String -> IO ()
> runLexer fileName = do
>   fileString <- readFile fileName
>   print (lexInit fileString)

> -- usada apenas para debugar a parte sintatica
> -- faz a analise sintatica de um arquivo, dado o nome do mesmo
> runParser :: String -> IO ()
> runParser fileName = do
>   fileString <- readFile fileName
>   print ((parserLF1 . lexInit) fileString)

> }

