{-
LINGUAGEM FUNCIONAL 1

MDULO MAIN:
  Mdulo principal, responsvel pela execuo do compilador

Equipe:
  Andr Wilson Brotto Furtado
  Gustavo Danzi de Andrade
  Monique Louise de Barros Monteiro

-}

module Main where

import Lexparser
import InterCodeMonad
import InterCodeGeneration
import FinalCodeGeneration
-- import Semantic
import Monad(when,unless)
import System

main :: IO ()
main = do
  args <- getArgs
  when (length args /= 1) (error "Uso: lf1c <nome do arquivo>.lf1")
  lf1c (head args)

lf1c :: String -> IO ()
lf1c file = do

    -- checa se o formato do arquivo  vlido (.lf1)
    when (fileFormatError file) (error "O compilador compila apenas arquivos .lf1!")

    -- l o diretrio do .NET framework
    baseDir <- readFile "dotnetpath.txt"
    
    -- l o arquivo
    fileString <- readFile file
    
    let tokens = lexInit fileString    -- faz a anlise lxica, criando a lista de tokens
        parseTree = parserLF1 tokens   -- faz a anlise sinttica, criando a parse tree
        prog = noExtension file        -- tira a extenso do arquivo (vai ser utilizado depois)

    -- gera o cdigo intermedirio, tambm a partir da parse tree e da tabela de smbolos
    let (intermediateCode,_) = applyInterCode (genInterCode parseTree) initInterCodeState
        finalCode = genFinalCode prog intermediateCode -- gera o cdigo final, a partir do intermedirio
        ilFile = changeExtension file		                                         -- gera o nome arquivo com extenso .asm
    -- escreve o cdigo intermedirio no arquivo inter.txt
    writeFile "inter.txt" (printFunctionsInterCode (functions intermediateCode)) 

    -- escreve o cdigo final no arquivo .il
    writeFile ilFile finalCode
    
    -- faz a compilao para il
    success <- system (baseDir ++ "ilasm " ++ ilFile ++ " /DLL")
    when (success /= ExitSuccess) (error "Erro na gerao da IL!")

    return ()


fileFormatError :: String -> Bool
fileFormatError file = fileFormatErrorAux (reverse file)

fileFormatErrorAux :: String -> Bool
fileFormatErrorAux ('1':'f':'l':'.':_) = False
fileFormatErrorAux _ = True

changeExtension ::  String -> String
changeExtension file = reverse (drop 3 (reverse file)) ++ "il"

noExtension :: String -> String
noExtension file = reverse $ drop 4 $ reverse file

printFunctionsInterCode :: [FunctionInfo] -> String
printFunctionsInterCode [] = []
printFunctionsInterCode (f:fs) = printInterCode (codeList f) ++ "\n\n" ++ (printFunctionsInterCode fs)

printInterCode :: [InterElement] -> String
printInterCode [] = []
printInterCode (i:is) = do
	show(i) ++ "\n" ++ printInterCode is

----------------------------------------------------------------------------------
--              INICIALIZAO DO ESTADO DO CDIGO INTERMEDIRIO                 --
----------------------------------------------------------------------------------
initInterCodeState :: InterCodeState
initInterCodeState = InterCodeState {
	nextLabel  = 0,
	renameFunctionsList  = [],      -- associa um identidicador de funo a um TermFunction
	functions = [],                 -- funes da aplicao
	functionContext	= []			   -- pilha cujo topo  a funo na qual o cdigo intermedirio deve ser gerado
}