module Main where

import Lexparser
import Semantic
import InterCodeMonad
import InterCodeGeneration
import FinalCodeGeneration
import Monad(when,unless)
import System

main :: IO ()
main = return ()

baseDir :: String
baseDir = "c:\\WINNT\\Microsoft.NET\\Framework\\v1.0.3705\\"

le2c :: String -> IO ()
le2c file = do

    -- checa se o formato do arquivo  vlido (.le2)
    when (fileFormatError file) (error "O compilador compila apenas arquivos .le2!")

    -- l o arquivo
    fileString <- readFile file

    let tokens = lexInit fileString    -- faz a anlise lxica, criando a lista de tokens
        parseTree = parserLE2 tokens   -- faz a anlise sinttica, criando a parse tree
        prog = noExtension file        -- tira a extenso do arquivo (vai ser utilizado depois)
    
    -- gera o cdigo intermedirio, tambm a partir da parse tree e da tabela de smbolos
    let (intermediateCode,result) = applyInterCode (genInterCode parseTree) initInterCodeState
    	finalCode = genFinalCode prog intermediateCode (getProgramType parseTree) result -- gera o cdigo final, a partir do intermedirio
        ilFile = changeExtension file		                                     -- gera o nome arquivo com extenso .asm
    
    -- escreve o cdigo intermedirio no arquivo inter.txt
    writeFile "inter.txt" (printInterCode2 (codeList intermediateCode))

    -- escreve o cdigo final no arquivo .il
    writeFile ilFile finalCode

    -- faz a compilao para il
    success1 <- system (baseDir ++ "ilasm " ++ ilFile)
    when (success1 /= ExitSuccess) (error "Erro na gerao da IL!")

    return ()


fileFormatError :: String -> Bool
fileFormatError file = fileFormatErrorAux (reverse file)

fileFormatErrorAux :: String -> Bool
fileFormatErrorAux ('2':'e':'l':'.':_) = False
fileFormatErrorAux _ = True

changeExtension ::  String -> String
changeExtension file = reverse (drop 3 (reverse file)) ++ "il"

noExtension :: String -> String
noExtension file = reverse $ drop 4 $ reverse file

printInterCode2 :: [InterElement] -> String
printInterCode2 [] = []
printInterCode2 (i:is) = do
	show(i) ++ "\n" ++ printInterCode2 is

----------------------------------------------------------------------------------
--              INICIALIZAO DO ESTADO DO CDIGO INTERMEDIRIO                 --
----------------------------------------------------------------------------------
initInterCodeState :: InterCodeState
initInterCodeState = InterCodeState {
	nextTemp   = 0,
	codeList   = [],
	renameList = [],
	dataTypes  = []
	}


